from unittest2 import TestCase
from aux.protocol.http.http import HTTP
import struct
import os
import re
from ..data.wsdl_raw_chunked_response import chunked_message, received_wsdl_message
from aux.protocol.transport import TCP_DEFAULT_FRAME_SIZE
from aux.protocol.http.transfer import ChunkedController

class FakeTransport(object):

    def __init__(self, message):
        self.fake_message = message
        self.bytes_read = 0
        
    def recv(self, nofchar=1024):
        buffer = ""
        for n in xrange(0, nofchar):
            if self.bytes_read >= len(self.fake_message):
                break
            else:
                buffer += self.fake_message[self.bytes_read]
            self.bytes_read += 1
        return buffer
    
    def close(self):
        pass

class HTTP_RECEIVE_LONG_WSDL_TEST(TestCase):
    
    def test_chunked_controller(self):
        transport = FakeTransport(chunked_message)
        inbuf = transport.recv(TCP_DEFAULT_FRAME_SIZE)
        inbuf = inbuf.split("\n")
        sl = inbuf[0]
        #Validate start-line and remove it from buffer
        re_startline = re.compile(r'^HTTP\/\d\.\d\s(\d{3})\s')
        tail_msg = "\n".join(inbuf[1:])

        status = int(re_startline.match(sl).groups()[0])

        re_headline = re.compile(r'^([A-Za-z\-]*)\s?\:\s?(.*)')
        headers = dict()
        body = ""
        t_lines = tail_msg.split("\r\n\r\n")
        if len(t_lines) == 1: #Servers can respond with \r\n or \n
            t_lines = tail_msg.split("\n\n")
        h_lines = t_lines[0].split("\n")
        line_counter = 0
        for line in h_lines:
            line_counter += 1
            if ":" in line:
                re_group = re_headline.match(line).groups()
                headers[re_group[0]] = re_group[1]
            else:
                break
        tail_msg = tail_msg[len(t_lines[0]):]

        self.assertEquals(headers.get('Transfer-Encoding'),
                          'chunked')
        Transfer = ChunkedController(headers, transport, tail_msg)
        body = Transfer.read()
        self.assertEqual(body,
                         received_wsdl_message)
        self.assertEqual(Transfer.chunks_in_stream,
                         ['1f09', '2000', '2000', '2000', '2000',
                          '2000', '2000', '2000', '2000', '2000',
                          '2000', '2000', '2000', '2000', '2000',
                          '2000', '2000', '2000', '2000', '2000',
                          '2000', '2000', '2000', '2000', '2000',
                          '1000', '81', '0'])

    def test_receive_long_wsdl(self):
        message = chunked_message

        http = HTTP()
        http.has_trace(True)
        response = http.receive(FakeTransport(message))
        self.assertEquals(response.status,
                          200)
        self.assertEqual(response.body,
                         received_wsdl_message)




from aux.protocol.transport import TCP_DEFAULT_FRAME_SIZE
import re


class NoContentController(object):
    def __init__(self, headers, transport, msg):
        self.headers = headers
        self.transport = transport
        self.msg = msg.lstrip()

    def read(self):
        return self.msg.rstrip()


class DefaultController(NoContentController):

    def read(self):
        raw_response = self.msg
        content_length = int(self.headers.get('Content-Length', 0))
        response = ""
        while 1:
            if content_length < 1:
                break
            if content_length > len(raw_response):
                raw_response += self.transport.recv()
            response += raw_response
            content_length -= len(raw_response)
            raw_response = ""
        return response.rstrip()


class ChunkedController(NoContentController):

    def __init__(self, headers, transport, msg):    
        super(ChunkedController, self).__init__(headers, transport, msg)
        self.chunks_in_stream = list()        

    def consume_buffer(self, i_chunk, raw_buffer):
        return raw_buffer[:i_chunk]
    
    def read(self, has_trace=False):
        re_chunk = re.compile(r'^([a-f|\d]{1,4})')
        re_chunk_block = re.compile(r'[\r\n]{0,2}[a-f|\d]{1,4}[\r\n]{0,2}', re.DOTALL)
        re_end_chunk = re.compile(r'^0\r\n\r\n0')
        re_single_end_chunk = re.compile(r'0[\r\n]{0,2}')
        raw_response = self.msg
        response = ""
        block = 0
        chunkcounter = 0
        i_chunk = 0

        while 1:
            if chunkcounter == 0:#When chunk has been consumed
                chunk = re_chunk.findall(raw_response[0:8])
                end_chunk = re_end_chunk.findall(raw_response[0:8])
                broken_end_chunk = re_single_end_chunk.findall(raw_response[0:8])
                if len(chunk) > 0:
                    self.chunks_in_stream.append(chunk[0])                    
                    i_chunk = int(chunk[0], 16)
                    chunkcounter = i_chunk
                    chunk_block = re_chunk_block.findall(raw_response)[0].replace(chunk[0], '')
                    nllength = len(chunk_block)
                    raw_response = raw_response[len(chunk[0])+nllength:]
                    # print end_chunk, broken_end_chunk
                    if i_chunk == 0 or len(end_chunk) > 0:
                        break
                    while i_chunk > len(raw_response):
                        raw_response += self.transport.recv(TCP_DEFAULT_FRAME_SIZE)
            # if len(broken_end_chunk) > 0:
            #     print raw_response[:-1*sum([len(e) for e in broken_end_chunk])]
            #     break
            
            response += self.consume_buffer(i_chunk, raw_response)
            raw_response = raw_response[i_chunk+1:].lstrip()
            chunkcounter -= i_chunk
        return response
    
def transferFactory(headers):
    content_length = headers.get('Content-Length', None)
    if content_length != None:
        if int(content_length) < 1:
            return NoContentController
    content_type = headers.get('Transfer-Encoding', None)
    if content_type != None:
        if 'chunked' in content_type.lower():
            return ChunkedController
    return DefaultController



print "%s is deprecated" % __file__
from aux.result import ResultCollector
from aux.backend import Backend

class ScriptEngine(object):
    def __init__(self):
        self.logging = None
        self.result_collector = None
        self.backend = None
        self.devices = {}

    def start(self):
        '''
        Initializes the script engine which in turn:

        - Starts the backend
        - Start result collection
        - Initialises logging

        '''
        self.logging = self.start_logging()
        self.result_collector = self.start_result_collector()
        self.backend = self.start_backend()

    def stop(self):
        self.stop_backend()
        self.stop_logging()
        return self.result_collector.get_results()

    def start_logging(self):
        import logging
        logging.basicConfig()
        return None

    def start_result_collector(self):
        result_collector = ResultCollector()
        return result_collector

    def start_backend(self):
        backend = Backend()
        backend.start()
        return backend

    def stop_backend(self):
        self.backend.stop()

    def stop_logging(self):
        pass

    def create_device(self, device_type, identifier, **extra_kwargs):
        if identifier in self.devices:
            return self.devices[identifier]
        device = device_type(identifier, self)
        self.devices[identifier] = device
        return device



# -*- coding: utf-8 -*-

from aux.system.base import BaseSystem

class Service(BaseSystem):
    def __init__(self, hostname):
        super(Service, self).__init__(hostname)
        self.hostname = hostname




from unittest2 import TestCase
from aux.protocol.http import HTTP, HTTPRequest
from ..util.mockhttpserver import MockHTTPServer


class HTTPProtocolTest(TestCase):

    def setUp(self):
        self.test_server = MockHTTPServer(port=8989)
        self.test_server.start_thread()
        self.headers = {'Host': 'a.a.a',
                        'User-Agent': 'Aux/0.1 (X11; Ubuntu; Linux x86_64; rv:24.0)',
                        'Accept':'text/html, application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                        'Accept-Language': 'en-US,en-q=0.5',
                        'Referer': 'http://abc.abc',
                        'Cache-Control': 'max-stale=0',
                        'Connection': 'Keep-Alive'
                        }
        
    def tearDown(self):
        self.test_server.stop()
    
    def xtest_connection(self):
        http = HTTP()
        url = 'http://127.0.0.1:8989'
        self.headers['Test-Controller'] = 'short_http_response'
        request = HTTPRequest(url,
                              {'method':'GET',
                               'headers': self.headers,
                               'data': 'fakedata'})
        response = http.send(request)
        print 'response: [', response, ']'
        self.assertTrue('200 OK' in response)
        self.assertTrue('<html>' in response)

        
    def xtest_handle_long_response(self):
        http = HTTP()
        url = 'http://127.0.0.1:8989'
        self.headers['Test-Controller'] = 'long_http_response'
        request = HTTPRequest(url,
                              {'method':'GET',
                               'headers': self.headers,
                               'data': 'fakedata'})
        response = http.send(request)
        print 'response: [', response, ']'
        #create a test mock handler http in backend



from aux.protocols import installers
from aux.device.base import Device

class NetworkDevice(Device):
    """Generic network device
    """
    def __init__(self, scriptengine, address, protocols):
        """
        @param scriptengine: see :Device.__init__:
        @param address: the network address of the device (usually hostname or ip)
        @param protocols: dictionary of protocol names and configurations
                          this Device should be configured with
        """
        Device.__init__(self, address, scriptengine)
        self.address = address

        for protocol, config in protocols.iteritems():
            # Find protocol installers for requested protocols and install on
            # this device with given configuration.
            installer = getattr(installers, protocol)
            installer(self, **config)



from aux.system.base import BaseSystem
from aux.authentication import BaseCredentials
from aux.api import ssh

class LinuxDevice(BaseSystem):
    def __init__(self, hostname):
        super(LinuxDevice, self).__init__(hostname)
        self.hostname = self.identifier
        self.ssh = ssh
        # self.ssh.set_hostname(self.identifier)




import sys        
import os
import shutil
import json
import logging
from optparse import OptionParser, OptionGroup
from aux import base_dir, systems_pool
from aux.system import get_system

DEFAULT_PROPERTIES_FILE = os.path.abspath(base_dir()+"/../"+"aux.properties")
DEFAULT_AUXLOG_FOLDER = "auxlogs"

def str2loglevel(option, opt_str, value ,parser):
    if value is not None:
        if "DEBUG"==value.upper():
            setattr(parser.values, option.dest, logging.DEBUG)
        elif "INFO"==value.upper():
            setattr(parser.values, option.dest, logging.INFO)
        elif "WARNING"==value.upper():
            setattr(parser.values, option.dest, logging.WARNING)
        elif "ERROR"==value.upper():
            setattr(parser.values, option.dest, logging.ERROR)            
        elif "CRITICAL"==value.upper():
            setattr(parser.values, option.dest, logging.CRITICAL)            
        else:
            setattr(parser.values, option.dest, logging.NOTSET)

def dir2abspath(option, opt_str, value ,parser):
    setattr(parser.values, option.dest, os.path.abspath(value))
            
class Configuration(object):
    def __init__(self):
        self.provision_optargs()

    def provision_optargs(self):
        usage = "usage: aux aux_file.py --option"
        parser = OptionParser(usage=usage)
        parser.add_option("--systems",
                          dest="systems",
                          type="string",
                          help="list of systems")
        parser.add_option("--config",
                          dest="configurationfile",
                          type="string")
        log_group = OptionGroup(parser, "Logging Options")
        log_group.add_option("-v", "--verbose",
                          dest="verbose",
                          action="store_true",
                          default=False,
                          help="verbose to console")
        log_group.add_option("--loglevel",
                          action="callback",
                          callback=str2loglevel,
                          dest="log_level",
                          type="string",
                          default=10,
                          help="[ NOTSET | ERROR | WARNING | INFO | DEBUG ]")
        log_group.add_option("--logconsolelevel",
                          action="callback",
                          callback=str2loglevel,
                          dest="log_console_level",
                          type="string",
                          default=None)
        log_group.add_option("--logfilelevel",
                          action="callback",                          
                          callback=str2loglevel,
                          dest="log_file_level",
                          type="string",
                          default=None)        
        log_group.add_option("--logdirectory",
                          action="callback",
                          callback=dir2abspath,   
                          dest="log_directory",
                          default=None,
                          type="string")
        log_group.add_option("--logsrv",
                          dest="log_server",
                          type="string")

        tools_group = OptionGroup(parser, "Auxiliary Tools")
        tools_group.add_option("--create_plugin",
                               dest="plugincreator",
                               type="string",
                               help="[ service, device, protocol ]")
        
        parser.add_option("--engine",
                          dest="engine_type",
                          type="string")
        
        parser.add_option_group(log_group)
        parser.add_option_group(tools_group)
        self.options, self.args = parser.parse_args()

    def load_default_properties(self):
        if self.options.configurationfile is None:
            configfilestr = os.path.abspath(os.path.join(os.path.expanduser("~"), ".aux/aux.properties"))
        else:
            configfilestr = self.options.configurationfile
        if not os.path.exists(configfilestr):
            configfilestr = DEFAULT_PROPERTIES_FILE
        fp = open(configfilestr, "r")
        file_configs = json.loads(fp.read())
        
        self.options.log_folder_name = DEFAULT_AUXLOG_FOLDER
        if self.options.log_server is None:
            self.options.log_server = file_configs.get('logging').get('resultServer')
        if self.options.log_directory is None:
            self.options.log_directory = file_configs.get('logging').get('logdirectory')
        if self.options.log_level is None:
            self.options.log_level = file_configs.get('logging').get('loglevel')
        if self.options.log_console_level is None:
            self.options.log_console_level = file_configs.get('logging').get('logconsolelevel')
        if self.options.log_file_level is None:
            self.options.log_file_level = file_configs.get('logging').get('logfilelevel')            
        if self.options.verbose is False:
            self.options.verbose = file_configs.get('logging').get('verbose')
            
    def set_systems(self):
        system_json = None
        if self.options.systems is not None:
            if '.json' in self.options.systems:
                url = os.path.abspath(self.options.systems)
                fp = open(url, 'r')
                try:                
                    system_json = json.loads(fp.read())
                except Exception, e:
                    print( "%s is not correctly formatted" % (self.options.systems))
                    print( e.message )                    
                fp.close()
            else:
                try:
                    system_json = json.loads(self.options.systems)
                except Exception, e:
                    print( "%s is not correctly formatted" % (self.options.systems))
                    print( e.message )
    
        if system_json is not None:
            for system in system_json:
                #populate pool of resources
                #TODO: handle if systems is not set to list
                systems_pool.append( system )
            
config = Configuration() if 'aux' in sys.argv[0] else None




from unittest2 import TestCase
from aux.engine import Engine
import time

class EngineTest(TestCase):

    def setUp(self):
        self.test_engine = Engine()

    def tearDown(self):
        del self.test_engine
    
    def xtest_start_stop_engine(self):
        self.assertFalse(self.test_engine.is_running() )
        self.test_engine.start()
        self.assertTrue(self.test_engine.is_running() )
        # time.sleep(2)
        self.test_engine.stop()
        self.assertFalse(self.test_engine.is_running() )
        





from socket import (AF_INET, SOCK_DGRAM, SOL_SOCKET,
                    SO_REUSEADDR, SHUT_RDWR)
from multiprocessing import Process
from ssl import wrap_socket
import ssl
import socket
import time
from aux.protocol.http import HTTPResponse, HTTPRequest
# from werkzeug import Response

#TODO: This server needs to be written using aux itself, listening service.
#but this is not implemented yet.

http_response ='''\
HTTP/1.1 200 OK
Date: Sun, 18 Oct 2009 08:56:53 GMT
Server: AuxMockHTTPServer (Unix)
Last-Modified: Sat, 20 Nov 2004 07:16:26 GMT
ETag: "10000000565a5-2c-3e94b66c2e680"
Accept-Ranges: bytes
Content-Length: 44
Connection: close
Content-Type: text/html
  
<html><body><h1>It works!</h1></body></html>
'''


def wsdl_app(environ, start_response):
    fake_wsdl_data = open("../data/geoipservice.asmx?WSDL").read()
    return """HTTP/1.1 200 OK
Server: mockhttpserver/1.5.4
Date: Mon, 10 Mar 2014 14:38:44 GMT
Content-Type: text/xml
Content-Length: %i
Transfer-Encoding: chunked
Connection: keep-alive\n\n%s""" % (len(fake_wsdl_data), fake_wsdl_data)

def soap_app(environ, start_response):
    return'''
HTTP/1.1 200 OK
Server: nginx/1.5.4'
Date: Wed, 12 Feb 2014 09:58:13 GMT
Content-Type: text/xml;charset=utf-8
Content-Length: 4734
Connection: keep-alive
Set-Cookie: JSESSIONID=oijoij6E0C7479C4CF531A5842241F47; Path=/; HttpOnly
X-Request-Received: 1392199092003
SOAPAction: ""

<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/">
<SOAP-ENV:Header/><SOAP-ENV:Body></SOAP-ENV:Body></SOAP-ENV:Envelope>
'''

def authenticator_app(environ, start_response):
    if environ.get('authScheme', None) == None:
        return True
    if 'Authorization: Basic' in environ.get('request', None):
        return True
    return False

def application(environ, start_response):
    # if not authenticator_app(environ, start_response):
    #     return "HTTP/1.1 403\n\n"
    if 'WSDL' in environ.get('request', None):
        return wsdl_app(environ, start_response)
    
    if 'SOAPAction' in environ.get('request', None):
        return soap_app(environ, start_response)
    return "HTTP/1.1 404\n\n"

def call_application(app, environ):
    body = []
    status_headers = [None, None]
    def start_response(status, headers):
        status_headers[:] = [status, headers]
        return body.append(status_headers)
    app_iter = app(environ, start_response)
    try:
        for item in app_iter:
            body.append(item)
    finally:
        if hasattr(app_iter, 'close'):
            app_iter.close()
    return status_headers[0], status_headers[1], ''.join(body)


class MockHTTPServer(object):
    def __init__(self, port=8989, verbose=False):
        self.port = port
        self.host = '127.0.0.1'
        self.__socket = socket.socket()
        self.__socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.__authScheme = None
        
    def start(self):
        self.__socket.bind((self.host, self.port))
        self.__socket.listen(5)
        while True:
            c, addr = self.__socket.accept()
            request = c.recv(4096)#TODO: read protocol transport
            print request
            response = call_application(application,
                                        {'request': request,
                                         'authScheme': self.__authScheme})
            c.send(response[2])
            c.close()

    def start_thread(self):
        self.p = Process(target=self.start)
        self.p.daemon = True
        self.p.start()
        time.sleep(.01)

    def stop(self):
        self.p.terminate()
        self.__socket.close()

    def set_authentication(self, authentication):
        self.__authScheme = authentication
        

class MockHTTPSServer(MockHTTPServer):
    def __init__(self, port=8443):
        self.parent = super(MockHTTPSServer, self)
        self.parent.__init__(port=port)
        self.__socket = socket.socket()
        self.__socket = socket.socket()
        self.__socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.__authScheme = None

    def start(self):
        self.__socket.bind((self.host, self.port))
        self.__socket.listen(5)
        while True:
            ssl_sock, addr = self.__socket.accept()
            sock = wrap_socket(
                ssl_sock,
                server_side=True,
                certfile='../data/certs/unit-test.crt',
                keyfile='../data/certs/unit-test.key',
                ssl_version=ssl.PROTOCOL_TLSv1)
            response = call_application(application,
                                        {'request': sock.read(),
                                         'authScheme': self.__authScheme})
            sock.send(response[2])
            sock.close()


    def stop(self):
        self.__socket.shutdown(SHUT_RDWR)
        self.parent.stop()


    def set_authentication(self, authentication):
        self.__authScheme = authentication

class Channel(object):

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.__socket = socket.socket()
        self.__socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        
    def start(self):
        self.__socket.bind((self.host, self.port))
        self.__socket.listen(5)
        # while True:
        #     c, addr = self.__socket.accept()
        #     request = c.recv(4096)#TODO: read protocol transport
        #     print request
        #     response = call_application(application,
        #                                 {'request': request,
        #                                  'authScheme': self.__authScheme})
        #     c.send(response[2])
        #     c.close()

    def stop(self):
        pass
    
        
class WebService(object):
    
    def __init__(self, port):
        self.scheme = 'http' # | https
        self.channel = Channel('127.0.0.1', 443)
        self.app = None

    def marshall_incoming_request(self, raw_request):
        return HTTPRequest('127.0.0.1', {})
        
    def call_app(self):
        request = self.marshall_incoming_request("POST /url")
        response = None
        try:
            response = self.app("environ", "start_response")
        except Exception, e:
            response = HTTPResponse(500, {})
        return response

    
class WebServer(object):
    def __init__(self):
        self.log = None
        self.services = []

    def load_service(self, service):
        self.services.append(service)

    def load_services(self, services=[]):
        for service in services:
            self.load_service(service)
        
    def start_service(self, service_name):
        for service in self.services:
            if service.name == service_name:
                service.channel.start()

    def stop_service(self, service_name):
        for service in self.services:
            if service.name == service_name:
                service.channel.stop()                

    def start(self):
        #start server
        for service in self.services:
            service.channel.start()

    def stop(self):
        #stop server
        for service in self.services:
            try:
                service.channel.stop()
            except Exception, e:
                service.channel.kill()

    def restart(self):
        #restart server
        self.stop()
        self.start()
    
        
    
#*******************************************************************
#OL' HEAP
#
# def handle_request(cls, request):
#     # print request
#     if cls._MockHTTPServer__authScheme != None:
#         if 'basic' in cls._MockHTTPServer__authScheme.lower() :
#             if not authenticate(request):
#                 return '''
# HTTP/1.1 401
# WWW-Authenticate: Basic realm="aux realm"
# Content-Type: text/xml;charset=utf-8
# Connection: keep-alive

# '''
#     if '__GET /basic_authenticated' in request:
#         return '''
# HTTP/1.1 403 OK

# basic auth'''
# #         https_response = '''\
# # HTTP/1.1 401
# # WWW-Authenticate: Basic realm="AUX-test"
# # '''
#     return http_response



from unittest2 import TestCase
from aux.logger import LogController


class LoggingTest(TestCase):

    def xtest_start_logging(self):
        pass
        #setting up log controller
        #logcontroller = LogController(config)

        #test write formats

        # log.info("This is a info message.")
        # log.debug("This is a debug message.")
        # log.error("This is an error message.")
        # log.warning("This is a warning message.")
        # log.critical("This is a critical message.")



from aux.protocol.soap import SOAPRequest, SOAPResponse
from lxml import objectify
from lxml import etree
from urlparse import urlparse
from aux.api import http
import os
import logging

log = logging.getLogger("protocol")

class WSDLPort(object):
    def __init__(self, e):
        self.e = e
        self.name = e.get('name')
        self.binding = e.attrib.get('binding')
        self.ext_element = e.getchildren()[0]

class WSDLPortType(object):
    def __init__(self, e):
        self.e = e
        self.name = self.e.get('name')
        self.operations = [WSDLOperation(o) for o in self.e.findall('%soperation' % WSDLClient.get_ns(self.e, 'wsdl'))]

#WIKI: Also reads as PortType in old definition        
class WSDLInterface(WSDLPortType):pass

class XMLComplexType(object):
    def __init__(self, e):
        self.e = e
        # print etree.tostring(e)
        # print 
        self.seq = self.e.findall('%ssequence' % WSDLClient.get_ns(self.e, 'xs'))
        elems = None
        if len(self.seq) > 0:
            elems = [e for e in self.seq[0].findall('%selement' % WSDLClient.get_ns(self.e, 'xs'))]
        if elems is None:
            self.elements = []
        else:
            self.elements = elems
            # for el in elems:
                # print el.get('name')
                # print el.get('type')
                # print etree.tostring(el)
                # print

            
class XMLSimpleType(object):
    def __init__(self, e):
        self.e = e
        
        
class WSDLElement(object):    #WIKI: XML Element SimpleType|ComplexType
    def __init__(self, e):
        self.e = e
        self.name = self.e.get('name')
        self.subelements = [XMLComplexType(se) for se in self.e.findall('%scomplexType' % WSDLClient.get_ns(self.e, 'xs'))]
        self.subelements.extend([XMLSimpleType(se) for se in self.e.findall('%ssimpleType' % WSDLClient.get_ns(self.e, 'xs'))])

class WSDLSchema(object):
    def __init__(self, e):
        self.e = e
        self.elements = [WSDLElement(elem) for elem in e.findall('%selement' % WSDLClient.get_ns(e, 'xs'))]

    
class WSDLTypes(object):
    def __init__(self, e):
        self.e = e
        self.schemas = [WSDLSchema(s) for s in self.e]
        

class WSDLMessagePart(object):
    def __init__(self, e):
        self.e = e
        if self.e is not None:
            self.name = e.get('name')
            self.element = e.attrib.get('element')
        
class WSDLMessage(object):
    def __init__(self, e):
        self.e = e
        if self.e is not None:
            self.name = self.e.get('name')
            self.part = WSDLMessagePart(self.e.find('%spart' % WSDLClient.get_ns(self.e,
                                                                           'wsdl')))

class WSDLOperation(object):
    def __init__(self, e):
        self.e = e
        self.name = self.e.get('name')
        self.input = WSDLMessage( self.e.find('%sinput' % WSDLClient.get_ns(self.e,
                                                                      'wsdl')) )
        self.output = WSDLMessage( self.e.find('%soutput' % WSDLClient.get_ns(self.e,
                                                                        'wsdl')) )

class WSDLBinding(object):
    def __init__(self, e):
        self.e = e
        self.name = e.get('name')
        self.type = e.get('type')
        self.soap_binding = e.find('%sbinding' % WSDLClient.get_ns(e, 'soap'))
        self.operations = [WSDLOperation(o) for o in e.findall('%soperation' % WSDLClient.get_ns(e, 'wsdl'))]
        #self.soap_operation = ?
        
    
class WSDLService(object):
    def __init__(self, e):
        self.name = e.get('name')
        ns = '{%s}' % e.nsmap.get('wsdl') if e.nsmap.get('wsdl') else '{%s}' % e.nsmap.get(None)
        self.documentation = e.find('%sdocumentation' % ns)
        self.ports = [WSDLPort(p) for p in e.findall('%sport' % ns)]
    

class WSDLDefinitions(object):
    def __init__(self, e):
        self.e = e
        # print etree.tostring(self.e)
        self.name = e.get('name')
        self.types = [WSDLTypes(t) for t in self.e.findall('%stypes' % WSDLClient.get_ns(self.e, 'wsdl'))]
        self.messages = [WSDLMessage(m) for m in self.e.findall('%smessage' % WSDLClient.get_ns(self.e, 'wsdl'))]
        porttype = self.e.find('%sportType' % WSDLClient.get_ns(self.e, 'wsdl'))
        self.portType = WSDLPortType(porttype) if porttype is not None else None
        binding = self.e.find('%sbinding' % WSDLClient.get_ns(self.e, 'wsdl'))
        self.binding = WSDLBinding(binding) if binding is not None else None
        service = self.e.find('%sservice' % WSDLClient.get_ns(self.e, 'wsdl'))
        self.service = WSDLService(service) if service is not None else None
        

class WSDLClient(object):

    def __init__(self, wsdl_url=None, wsdl_data=None):
        #WIKI: descriptions is often called definitions.
        self.wsdl_data = wsdl_data
        self.__api_sources = list()
        self.definitions = dict()

    def set_api_source(self, new_api_source_dsn):
        self.__api_sources.append(new_api_source_dsn)

    def get_api_sources(self):
        return self.__api_sources

    def update_api(self):
        if len(self.get_api_sources()):
            for source in self.get_api_sources():
                self.load_wsdl(source.split('/')[-1].replace('.wsdl',''),
                               self.get_proxy(source))
        else:
            self.load_wsdl('wsdl_data', wsdl_data=self.wsdl_data)

    @classmethod
    def get_ns(cls, element, namespace):
        return '{%s}' % element.nsmap.get(namespace) if element.nsmap.get(namespace) else '{%s}' % element.nsmap.get(None)
    
    def load_wsdl(self, definition_name, wsdl_url=None, wsdl_data=None):
        wsdl_string = None
        if wsdl_url is not None:
            wsdl_string = http.get(wsdl_url,
                                   headers=self.headers).body
        elif wsdl_data is not None:
            wsdl_string = wsdl_data
        if wsdl_string is not None:
            resource = etree.XML(wsdl_string)
            self.definitions[definition_name] = WSDLDefinitions(resource)
            
    def __getattr__(self, attr):
        class WSDLo(object):
            def __init__(self, self_instance, definition):
                self.instance = self_instance
                self.definition = definition
                
            def __getattr__(self, attr):
                self.operation = [o for o in self.definition.portType.operations if o.name==attr][0]
                return self.operation_wrapper

            def operation_wrapper(self, kwargs):
                data_types = dict()
                for t in [ts for ts in self.definition.types]:
                    for s in [sc for sc in t.schemas]:
                        for e in [el for el in s.elements]:
                            if self.operation.input.name==e.name:
                                for subelm in e.subelements:
                                    for el in subelm.elements:
                                        data_types[el.get('name')] = el.get('type')
                soap_types = list()
                for key in kwargs.keys():
                    if key in data_types.keys():
                        soap_types.append( "<kcen:%s>%s</kcen:%s>" % (key,
                                                            kwargs[key],
                                                            key) )
                if 'http' in self.definition.binding.soap_binding.get('transport'):
                    transport = 'http://'
                elif 'https' in self.definition.binding.soap_binding.get('transport'):
                    transport = 'https://'
                request_url = os.path.join(transport,
                                           self.definition.service.ports[0].ext_element.get('location'),
                                           self.definition.portType.name)                        
                soap_body = """<kcen:%s>%s</kcen:%s>""" % (self.operation.input.name,
                                                           "".join(soap_types),
                                                           self.operation.input.name)
                soaprequest = SOAPRequest(request_url, http.basic(self.instance.credentials), soap_body)
                return soaprequest.send()
            
        definition = self.definitions.get(attr, None)
        # print definition.service.name
        if definition is not None:
            return WSDLo(self,
                         self.definitions.get(attr))
        else:
            emsg = "%s object has no attribute '%s'" % (self.__class__.__name__,
                                                        attr)
            raise AttributeError(emsg)

            



import paramiko


class SSHMessage(object):
    def __init__(self):
        self.packet_length
        self.padding_length
        self.payload
        self.random_padding

#RFC4253
class SSHClient(object):
    credentials = None
    
    def __init__(self):
        self.hostname = None
        self.__connection = paramiko.SSHClient()
        self.__connection.load_system_host_keys()
        self.__connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    def set_hostname(self, hostname):
        self.hostname = hostname
        
    def set_credentials(self, credentials):
        self.credentials = credentials
    
    def cmd(self, command_msg):
        self.__connection.connect(self.hostname,
                                  username=self.credentials[0],
                                  password=self.credentials[1])
        stdin, stdout, stderr = self.__connection.exec_command(command_msg)
        return stdout.read()

#initial handshake

#resilliency signaling

#data (re)transmission



#[1] SSH Connection Protocol

#[2] SSH User Authentication Protocol

#[3] SSH Transport Protocol


"""
      uint32    packet_length
      byte      padding_length
      byte[n1]  payload; n1 = packet_length - padding_length - 1
      byte[n2]  random padding; n2 = padding_length
      byte[m]   mac (Message Authentication Code - MAC); m = mac_length

      packet_length
         The length of the packet in bytes, not including 'mac' or the
         'packet_length' field itself.

      padding_length
         Length of 'random padding' (bytes).

      payload
         The useful contents of the packet.  If compression has been
         negotiated, this field is compressed.  Initially, compression
         MUST be "none".

      random padding
         Arbitrary-length padding, such that the total length of
         (packet_length || padding_length || payload || random padding)
         is a multiple of the cipher block size or 8, whichever is
         larger.  There MUST be at least four bytes of padding.  The
         padding SHOULD consist of random bytes.  The maximum amount of
         padding is 255 bytes.

      mac
         Message Authentication Code.  If message authentication has
         been negotiated, this field contains the MAC bytes.  Initially,
         the MAC algorithm MUST be "none".
"""




class Promise(object):
    def __init__(self):
        self.callbacks = list()

    def addCallbacks(self, callback):
        self.callbacks.append(callback)

        
class Actor(object):
    pass

class Reactor(Actor):
    pass

class Proactor(Actor):
    pass

class Coactor(Actor):
    """
    A Coactor has a realtime queue and an async queue.
    Reactor and Proactor combined.
    """
    pass



from aux.protocol.soap.wsdl import WSDLDefinitions
from aux.api import http
from lxml import etree
import os


class WebServiceNotFoundError(Exception):pass


class WebService(object):

    def __init__(self, source, client):
        self.source = source
        self.client = client
        # print self.source


class WSDLService(WebService):

    def __init__(self, source, client):
        super(WSDLService, self).__init__(source, client)
        wsdl_response = http.get(self.client.get_proxy(self.source),
                                 headers=self.client.headers)
        # print wsdl_response.body
        resource = etree.XML(wsdl_response.body)
        self.definitions = WSDLDefinitions(resource)
        # print "source", self.source
        
    @classmethod
    def get_ns(cls, elmenet, namepspace):
        return '{%s}' % element.nsmap.get(namespace) if element.nsmap.get(namespace) else '{%s}' % element.nsmap.get(None)
        
        
class RESTService(WebService):
    
    def __init__(self, source, client):
        super(RESTService, self).__init__(source, client)        


def webservice_factory(source, client):
    if source.endswith('.wsdl'):
        return WSDLService(source, client)
    elif source.endswith('.json'):
        return RESTService(source, client)
    raise WebServiceNotFoundError("No handler for %s" % source)
    

class WSClient(object):

    def __init__(self, service):
        self.service = service
        self.__api_sources = list()
        self.webservices = dict()
        self.prefix = self.service.PREFIX
        self.SCHEME = self.service.get_scheme        
        self.headers = {'Host':self.service.hostname,
                        'User-Agent': 'Aux/0.1 (X11;Ubuntu;Linux x86_64;rv:24.0)',
                        'Accept':'*/*'}        


    def get_proxy(self, path, prefix=None):
        if prefix == None:
            prefix = self.prefix
        return "%s%s%s%s" % (self.SCHEME(),
                             self.service.hostname,
                             prefix,
                             path)
    
    def set_api_source(self, new_api_source_dsn):
        self.__api_sources.append(new_api_source_dsn)

    def get_api_sources(self):
        return self.__api_sources

    def update_api(self):
        try:
            for source in self.get_api_sources():
                source_name = os.path.basename(source).replace('.wsdl','')
                self.webservices[source_name] = webservice_factory(source, self)
        except WebServiceNotFoundError, e:
            print e.message

    def set_credentials(self, credentials):
        self.credentials = credentials
        # self.headers.update(http.basic(self.credentials))
        # log.debug("WSDL credentials changed")

    def get_credentials(self):
        return self.credentials

    
    def __getattr__(self, attr):
        # print 'attr', attr
        definition = self.webservices.get(attr, None)
        if definition is not None:
            print definition
            return definition
        else:
            emsg = "%s object has no attribute '%s'" % (self.__class__.__name__,
                                                        attr)
            raise AttributeError(emsg)
            
        #     return WSDLo(self,
        #                  self.definitions.get(attr))
        # else:
        #     emsg = "%s object has no attribute '%s'" % (self.__class__.__name__,
        #                                                 attr)
        #     raise AttributeError(emsg)



import paramiko
from aux.protocol.http import HTTPClient
from aux.protocol.ssh import SSHClient
from aux.internals.credentials import Credentials

# def run(engine, func, *args, **kwargs):
#     engine.start()
#     try:
#         func(*args, **kwargs)
#     finally:
#         results = engine.stop()
#     return results

ssh = SSHClient()
http = HTTPClient()









from aux.engine.actor import (Reactor, Proactor, Coactor, NoActorFoundError)
import sys
import traceback

import logging

runtime_log = logging.getLogger('runtime')

class Engine(object):

    def __init__(self, Actor=Reactor, configuration=None):
        self.actor = Actor("Engine")
        
    def setup(self, kwargs):
        self.actor.setup()

    def add_callback(self, callback):
        self.actor.add_callback(callback)

    def remove_callback(self, callback):
        self.actor.remove_callback(callback)
        
    def is_running(self):
        return self.actor.is_running()
        
    def start(self):
        try:
            # self.actor.start()
            # runtime_log.info("Engine started")
            runtime_log.debug("Engine start not implemented")            
        except:
            print traceback.print_exc(file=sys.stdout)

    def stop(self):
        
        try:
            # self.actor.stop()
            # runtime_log.info("Engine stopped")
            runtime_log.debug("Engine stop not implemented")
        except Exception:
            print traceback.print_exc(file=sys.stdout)
        finally:
            if self.is_running():
                pass

def engine_factory(engine_type, config):
    if engine_type.upper()=='REACTOR':
        return Engine(Reactor, configuration=config)
    elif engine_type.upper()=='PROCATOR':
        return Engine(Proactor, configuration=config)
    elif engine_type.upper()=='COACTOR':
        return Engine(Coactor, configuration=config)
    else:
        raise NoActorFoundError(engine_type)



# DEPRECATED:
# from unittest2 import TestCase
# from aux.service.servicefactory import service_factory
# from aux.protocols.http import HTTPRequest as Request


# class FactoryServiceTest(TestCase):

#     def xtest_http_service(self):

#         service = service_factory('https')
#         request = Request({'method': 'GET',
#                            'headers': ['Content-Type: application/json',
#                                        'Accept-Type: application/json'],
#                            'data': "Test"
#                            })
#         service.get(request)



# -*- coding: utf-8 -*-
# from aux.system.base import BaseSystem
from linux import LinuxDevice


#TODO: obviously this is just thrown together

# class Device(BaseSystem):
#     def __init__(self, hostname):
#         super(Service, self).__init__(hostname)
#         self.hostname = hostname



import os
import re

#TODO: use mimeparse but keep factory pattern

class DefaultController(object):

    def __init__(self, disposition, raw_body):
        self.data = raw_body

    def handle(self):
        return self.data

    
class ZIPController(object):

    def __init__(self, disposition, raw_body):
        self.disposition = disposition
        self.data = raw_body

    def handle(self):
        re_filename = re.compile(r'attachment;\s?filename="(.*\.zip)"')
        tmp_dir = "/tmp/aux"
        filename = re_filename.findall(self.disposition)[0]
        filepath = os.path.join(tmp_dir, filename)
        if not os.path.exists(tmp_dir):
            os.mkdir(tmp_dir)
        fp = open(filepath, 'w')
        fp.write(self.data)
        fp.close()
        return filepath


def mimeFactory(headers):
    content_type = headers.get('Content-Type', None)
    if content_type != None:
        if 'application/zip' in content_type:
            return ZIPController
    return DefaultController



class BaseCredentials(object):
    def __init__(self, ip=None, username=None, password=None, port=None):
        self.ip = ip
        self.username = username
        self.password = password
        self.port = port



from unittest2 import TestCase
from aux.protocol.soap.wsdl import WSDLClient


class WSDLTest(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_basic_wsdl_descriptions_no_attrib(self):
        wsdl_data = "<definitions></definitions>"
        wsdl_object = WSDLClient(wsdl_data = wsdl_data)
        wsdl_object.update_api()
        self.assertEquals(None, wsdl_object.definitions.get('wsdl_data').name)

    def test_basic_wsdl_descriptions_called_definitions(self):
        wsdl_data = '<definitions name="TestService"></definitions>'
        wsdl_object = WSDLClient(wsdl_data = wsdl_data)
        wsdl_object.update_api()
        self.assertEquals('TestService', wsdl_object.definitions.get('wsdl_data').name)

    def test_basic_wsdl_descriptions(self):
        wsdl_data = '<descriptions name="TestService"></descriptions>'
        wsdl_object = WSDLClient(wsdl_data = wsdl_data)
        wsdl_object.update_api()
        self.assertEquals('TestService', wsdl_object.definitions.get('wsdl_data').name)

    def xtest_basic_wsdl_service(self):
        #TODO: might need tag closure preprocessor
        wsdl_data = """
<descriptions name=\"TestService\" 
   targetNamespace="http://www.examples.com/wsdl/HelloService.wsdl"
   xmlns="http://schemas.xmlsoap.org/wsdl/"
   xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/"
   xmlns:tns="http://www.examples.com/wsdl/HelloService.wsdl"
   xmlns:xsd="http://www.w3.org/2001/XMLSchema">
   <service name="HelloService">
      <documentation>WSDL File for HelloService</documentation>
      <port binding="tns:Hello_Binding" name="Hello_Port">
         <soap:address
            location="http://www.examples.com/SayHello/"/>
      </port>
   </service>
</descriptions>"""
        wsdl_object = WSDLClient(wsdl_data = wsdl_data)
        service = wsdl_object._w_services[0]
        self.assertEquals("HelloService", service.name)
        self.assertEquals("tns:Hello_Binding", service.ports[0].binding)
        self.assertEquals("Hello_Port", service.ports[0].name)
        self.assertEquals("%saddress" % WSDL.get_ns(service.ports[0].ext_element, "soap"),
                          service.ports[0].ext_element.tag )
        self.assertEquals("http://www.examples.com/SayHello/",
                          service.ports[0].ext_element.attrib.get('location'))

    def xtest_wsdl_service(self):
        wsdl_data = """
<wsdl:definitions xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/" xmlns:sch2="http://auxiliary.protojour.com/schemas/common" xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/" targetNamespace="http://auxiliary.protojour.com/schemas/test">
  <wsdl:types>
    <xs:schema xmlns="http://auxiliary.protojour.com/schemas/test" 
               xmlns:ns0="http://auxiliary.protojour.com/schemas/common" 
               xmlns:xs="http://www.w3.org/2001/XMLSchema" 
               attributeFormDefault="unqualified" 
               elementFormDefault="qualified" 
               targetNamespace="http://auxiliary.protojour.com/schemas/test">
      <xs:import namespace="http://auxiliary.protojour.com/schemas/common"/>
      <xs:element name="ListSomethingRequest">
        <xs:annotation>
          <xs:documentation>Request a list of all available code types that can be ordered.</xs:documentation>
        </xs:annotation>
        <xs:complexType/>
      </xs:element>
      <xs:element name="ListSomethingResponse">
        <xs:complexType>
          <xs:sequence>
            <xs:element maxOccurs="unbounded" minOccurs="0" name="codePattern">
              <xs:complexType>
                <xs:attribute name="id" type="xs:string" use="required">
                  <xs:annotation>
                    <xs:documentation>The ID of the code pattern</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="description" type="xs:string" use="required">
                  <xs:annotation>
                    <xs:documentation>A text description of the code pattern</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="minLength" type="xs:positiveInteger" use="required">
                  <xs:annotation>
                    <xs:documentation>The minimum code length that can be ordered</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="defaultLength" type="xs:positiveInteger" use="required">
                  <xs:annotation>
                    <xs:documentation>The default code length of this code pattern</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="maxLength" type="xs:positiveInteger" use="required">
                  <xs:annotation>
                    <xs:documentation>The maximum code length allowed for this code pattern</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="caseSensitive" type="xs:boolean" use="required">
                  <xs:annotation>
                    <xs:documentation>Whether this code is case sensitive or not</xs:documentation>
                  </xs:annotation>
                </xs:attribute>
              </xs:complexType>
            </xs:element>
          </xs:sequence>
          <xs:attribute name="defaultCase" type="CodeCase" use="required">
            <xs:annotation>
              <xs:documentation>The default code case if the code is case-insensitive</xs:documentation>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="InitOrderRequest">
        <xs:annotation>
          <xs:documentation>Create a new empty order containing zero codes. Use [expandOrder] to request codes for the order.</xs:documentation>
        </xs:annotation>
        <xs:complexType>
          <xs:sequence>
            <xs:element minOccurs="0" name="type" type="OrderType">
              <xs:annotation>
                <xs:documentation>Deprecated! Do not use</xs:documentation>
              </xs:annotation>
            </xs:element>
            <xs:element default="false" minOccurs="0" name="sectioned" type="xs:boolean">
              <xs:annotation>
                <xs:documentation>Should this order be split into section. Usaually used only with pre-printed media</xs:documentation>
              </xs:annotation>
            </xs:element>
            <xs:element minOccurs="0" name="description" type="xs:string">
              <xs:annotation>
                <xs:documentation>An optional description of the order</xs:documentation>
              </xs:annotation>
            </xs:element>
            <xs:element minOccurs="0" name="productId" type="xs:int">
              <xs:annotation>
                <xs:documentation>An optional product ID. Has to match a previously registered product ID, see [listProducts]</xs:documentation>
              </xs:annotation>
            </xs:element>
            <xs:element minOccurs="0" name="gtin" type="xs:string">
              <xs:annotation>
                <xs:documentation>An optional product GTIN. Has to match a previously registered product's GTIN, see [listProducts]</xs:documentation>
              </xs:annotation>
            </xs:element>
            <xs:element minOccurs="0" name="metadataSchema" type="xs:string">
              <xs:annotation><xs:documentation>An optional meta-data schema ID. Has to match a previously registered schema ID, see [listMetadataSchemas]</xs:documentation>
              </xs:annotation>
            </xs:element>
          </xs:sequence>
        </xs:complexType>
      </xs:element>
      <xs:element name="InitOrderResponse">
        <xs:complexType>
          <xs:sequence>
            <xs:element name="orderId" type="xs:long">
              <xs:annotation>
                <xs:documentation>The ID of the created order</xs:documentation>
              </xs:annotation>
            </xs:element>
          </xs:sequence>
        </xs:complexType>
      </xs:element>
    </xs:schema>
  </wsdl:types>
  <wsdl:message name="ListSomethingRequest">
    <wsdl:part element="tns:ListSomethingRequest" name="ListSomethingRequest">
    </wsdl:part>
  </wsdl:message>
  <wsdl:message name="ListSomethingResponse">
    <wsdl:part element="tns:ListSomethingResponse" name="ListSomethingResponse">
    </wsdl:part>
  </wsdl:message>
  <wsdl:message name="InitOrderRequest">
    <wsdl:part element="tns:InitOrderRequest" name="InitOrderRequest">
    </wsdl:part>
  </wsdl:message>
  <wsdl:message name="InitOrderResponse">
    <wsdl:part element="tns:InitOrderResponse" name="InitOrderResponse">
    </wsdl:part>
  </wsdl:message>

  <wsdl:portType name="auxiliary">
    <wsdl:operation name="ListSomething">
      <wsdl:input message="tns:ListSomethingRequest" name="ListSomethingRequest">
      </wsdl:input>
      <wsdl:output message="tns:ListSomethingResponse" name="ListSomethingResponse">
      </wsdl:output>
    </wsdl:operation>
    <wsdl:operation name="InitOrder">
      <wsdl:input message="tns:InitOrderRequest" name="InitOrderRequest">
      </wsdl:input>
      <wsdl:output message="tns:InitOrderResponse" name="InitOrderResponse">
      </wsdl:output>
    </wsdl:operation>
  </wsdl:portType>

  <wsdl:binding name="AuxiliarySoap11" type="tns:auxiliary">
    <soap:binding style="document" transport="http://schemas.xmlsoap.org/soap/http"/>
    <wsdl:operation name="ListSomething">
      <soap:operation soapAction=""/>
      <wsdl:input name="ListSomethingRequest">
        <soap:body use="literal"/>
      </wsdl:input>
      <wsdl:output name="ListSomethingResponse">
        <soap:body use="literal"/>
      </wsdl:output>
    </wsdl:operation>

    <wsdl:operation name="InitOrder">
      <soap:operation soapAction=""/>
      <wsdl:input name="InitOrderRequest">
        <soap:body use="literal"/>
      </wsdl:input>
      <wsdl:output name="InitOrderResponse">
        <soap:body use="literal"/>
      </wsdl:output>
    </wsdl:operation>
  </wsdl:binding>

  <wsdl:service name="auxiliaryService">
    <wsdl:documentation>Hello</wsdl:documentation>
    <wsdl:port binding="tns:AuxiliarySoap11" name="AuxiliarySoap11">
      <soap:address location="http://127.0.0.1/aux/auxiliary-ws"/>
    </wsdl:port>
  </wsdl:service>
</wsdl:definitions>
"""
        wsdl_object = WSDLClient(wsdl_data = wsdl_data)
        #message
        print 'debug', wsdl_object.definitions
        self.assertEquals("ListSomethingRequest",
                          wsdl_object.definitions[0].messages[0].name)
        self.assertEquals("ListSomethingResponse",
                          wsdl_object._w_messages[1].name)
        self.assertEquals("InitOrderRequest",
                          wsdl_object._w_messages[2].name)
        self.assertEquals("InitOrderResponse",
                          wsdl_object._w_messages[3].name)
        #service
        self.assertEquals("AuxiliarySoap11",
                          wsdl_object._w_services[0].ports[0].name)
        self.assertEquals("http://127.0.0.1/aux/auxiliary-ws",
                          wsdl_object._w_services[0].ports[0].ext_element.get('location'))
        #binding
        print wsdl_object.ListSomething()
        # self.assertEquals("Service



from unittest2 import TestCase
# from aux.protocol.rest.jsondl import JSONDL
from aux.protocol.soap.wsdl import WSDLClient
from wsdldatatmp import ssp_wsdl

WSDL_MOCK = '''<wsdl:definitions name="TestService"
   targetNamespace="http://www.examples.com/wsdl/HelloService.wsdl"
   xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/"
   xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/"
   xmlns:tns="http://www.examples.com/wsdl/HelloService.wsdl"
   xmlns:xsd="http://www.w3.org/2001/XMLSchema">
   <wsdl:types>
     <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">

     </xs:schema>
   </wsdl:types>
   <wsdl:message name="GetTestRequest">
     <wsdl:part element="tns:GetTest" name="GetTestRequest"></wsdl:part>
   </wsdl:message>
   <wsdl:portType name="testsomething">
      <wsdl:operation name="GetTest">

      </wsdl:operation>
   </wsdl:portType>
   <wsdl:service name="TestService">
      
   </wsdl:service>
</wsdl:definitions>
'''


JSONDL_MOCK = '''
{"api": {},
 "types": {}
}
'''


class TestWSDLObject(TestCase):
    def xtest_get_test_request(self):
        wsdl = WSDLClient(wsdl_data=ssp_wsdl)




# DEPRECATED: 
# from base64 import b64encode

# class Credentials(object):
#     def __init__(self, username, password):
#         self.username = username
#         self.password = password

#     def get_as_basic_auth(self):
#         return "Basic %s" % b64encode(b'%s%s' % (self.username, self.password))


# class BaseService(object):
#     _connection = None

# class HTTPService(BaseService):

    
#     def _head(self, request):
#         pass
    
#     def _post(self, request):
#         pass

#     def _delete(self, request):
#         pass

#     def get(self, request):
#         if request.method == 'post':
#             return self._post(request)
#         elif request.method == 'head':
#             return self._head(request)
#         elif request.method == 'delete':
#             return self._delete(request)


# class HTTPSService(HTTPService):
#     def __init__(self, args):
#         self.url = args.get('url', None)

# class ServiceFactory(object):
#     def __init__(self):
#         pass

#     def get_service(self, service_type, args):
#         if 'http' == service_type:
#             return HTTPService(args)
#         elif 'https' == service_type:
#             return HTTPSService(args)
#         return None


# def service_factory(servicetype, args={}):
#     factory = ServiceFactory()
#     return factory.get_service(servicetype, args)









class Callback(object):
    def __init__(self, callback_fn):
        pass

class Failure(object):
    pass

class Promise(object):

    def __init__(self):
        self.__callbacks = list()

    def add_callback(self, callbacks):
        self.__callbacks.append(callbacks)






from twisted.internet import reactor, threads
import threading
import functools

import aux.protocol as protocol_module

class Backend(object):
    def __init__(self):
        self.thread = None
        self.reactor = reactor
        self.event = threading.Event()
        self.protocols = protocols_module

    def start(self):
        self.thread = threading.Thread(name='BackendThread',
                                       target=self.start_reactor)
        self.thread.start()
        #The event.set is called when the reactor
        #is completely initialized.
        self.event.wait()

    def stop(self):
        self.reactor.callFromThread(self.reactor.stop)
        while self.thread.is_alive():
            # Do not just do .join() as this will block the mainthread
            # in such a way that C-c will not work.
            self.thread.join(timeout=0.01)

    def start_reactor(self):
        self.reactor.callWhenRunning(lambda: self.event.set())
        self.reactor.run(installSignalHandlers=0)

    def make_proxy(self, obj):
        if isinstance(obj, Proxy):
            raise AssertionError('Wrapping a Proxy in a Proxy will deadlock')
        return Proxy(obj)

class Proxy(object):
    def __init__(self, wrapped_obj):
        self.__dict__['wrapped_obj'] = wrapped_obj

    def __getattr__(self, attr):
        if attr in ['wrapped_obj']:
            return self.__dict__['wrapped_obj']
        if hasattr(self.wrapped_obj, attr):
            attr = getattr(self.wrapped_obj, attr)
            if callable(attr):
                return self.create_blocking_wrapper(attr)
            return attr
        raise KeyError('%s does not exist in %s' % (attr, self))

    def __setattr__(self, attr, value):
        setattr(self.wrapped_obj, attr, value)

    def create_blocking_wrapper(self, callable_):
        @functools.wraps(callable_)
        def _blocked(*args, **kwargs):
            return threads.blockingCallFromThread(reactor,
                                                  callable_,
                                                  *args,
                                                  **kwargs)
        return _blocked



from unittest2 import TestCase
from aux.protocols.http import HTTPResponse
from .mockhttpserver import WebService, WebServer


class MockHTTPServerTest(TestCase):

    """
    WebService
     - WebService with no app returns 500
     - WebService with success app returns 200
     - 
    """
    
    def test_no_app_defined(self):
        ws = WebService()
        response = ws.call_app()
        self.assertTrue(response.status == 500)

    def test_app_success_defined(self):
        ws = WebService()
        def test_app_ok(environ, start_response):
            return HTTPResponse(200, {'body': 'OK'})
        ws.app = test_app_ok
        response = ws.call_app()
        self.assertTrue(response.status == 200)

    def test_WebService_channel_listener(self):
        pass #Channel

    """
    WebServer
     - Holds a list of services
     - enables service channel listener
     - spawns request handler thread    
    """
    def test_WebServer_list_of_services(self):
        ws1 = WebService()
        ws2 = WebService()
        def test_app_ok(environ, start_response):
            return HTTPResponse(200, {'body': 'OK'})        
        ws1.app = test_app_ok
        ws2.app = test_app_ok
        server = WebServer()
        server.load_service(ws1)
        server.load_service(ws2)
        for service in server.services:
            self.assertTrue(service.call_app().status == 200)

    def test_WebServer_start_server(self):
        ws = WebService()
        def test_app_ok(enviorn, start_response):
            return HTTPResponse(200, {'body': 'OK'})
        ws.app = test_app_ok
        server = WebServer()
        server.load_services([ws, ])
        server.start()
            



from base64 import b64encode

class BasicAuthenticator(object):

    def __init__(self, credentials):
        self.credentials = credentials

    def __call__(self):
        return {"Basic":"%s" % b64encode(b'%s%s' % (self.credentials.username,
                                                     self.credentials.password))}



    




class BaseActor(object):

    def __init__(self, name):
        self.__is_running = False
        self.name = name

    def is_running(self):
        return self.__is_running

    def set_running(self, state):
        self.__is_running = state
        
    def start(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()





from multiprocessing import Process, Pipe
from aux.engine.actor.base import BaseActor
import select
        
class Reactor(BaseActor):

    def __init__(self, name, looper="select"):
        self.parent = super(Reactor, self)
        self.parent.__init__(name)
        self.callbacks = list()
        self.shouldStop = False
        self.p1_in, self.p1_out = Pipe()
        self.p2_in, self.p2_out = Pipe()
        self.select_timeout = 0.5

    def add_callback(self, callback):
        self.callbacks.append(callback)
        
    def mainloop(self, select_time_out, p_in, p_out):
        should_run = True
        reads = list()
        writes = list()
        exceptions = list()
        
        while ( should_run ):
            data =  p_in.recv()
            if 'should_run=False' in data:
                should_run = False
            select.select(reads,
                          writes,
                          exceptions,
                          self.select_timeout)
            # p_out.send("reactor send")
            # should_run = False
        
    def start(self):
        reactor_process = Process(target=self.mainloop,
                                  args=(self.select_timeout,
                                        self.p1_out,
                                        self.p2_in))
        reactor_process.daemon = True
        reactor_process.start()
        self.parent.set_running(True)
        
    def stop(self):
        self.p1_in.send('should_run=False')
        print self.p2_out.recv()
        self.parent.set_running(False)





class Credentials(object):
    
    def __init__(self, username, password):
        self.username = username
        self.password = password

    def __call__(self):
        return (self.username, self.password)
        
    def __repr__(self):
        return '(\'%s\', ***)' % (self.username)
        





from colorama import Fore, Back, Style
import os
import logging
from aux import logcontroller

def as_role(method):
    return method

def as_user(method):
    return method


def expected(expected_result,
             actual_result,
             service="",
             note=None,
             error_msg=None):
    log = logging.getLogger('script')

    if actual_result != expected_result:
        # color_s = Fore.RED + "Error: %s" + Fore.RESET
        color_s = " Err: %s"
        log.error(color_s % service)
        log.error(" Actual:%s != %s:Expected" % (actual_result, expected_result))
        logcontroller.summary['success'] = False        
        if note is not None:
            log.info("   : %s" % (note))
        if error_msg is not None:
            log.error("   : %s" % (error_msg))
    else:
        # color_s = Fore.GREEN + "OK: %s" + Fore.RESET
        color_s = " OK : %s"
        log.info(color_s % service)
        if note is not None:
            log.info("   : %s" % (note))



import base64
import logging

log = logging.getLogger("systems")

class Authentication(object):
    #TODO: this responsibility should be moved out of base system definition
    class BasicAuth(object):
        basic = None
        def __call__(self, json_data):
            self.basic = json_data.get('base64', None)
            if self.basic is None:
                self.basic = base64.b64encode("%s:%s" % (json_data.get('username'),
                                                         json_data.get('password') ))
                log.debug("Basic authentication credentials changed")
        def __repr__(self):
            return self.basic

        def dict(self):
            return {"Authorization" : "Basic %s" % self.basic}
        
    class SSHAuth(object):
        username = None
        password = None
        cert = None
        
        def __call__(self, json_data):
            self.username = json_data.get('username')
            self.password = json_data.get('password')

        def __repr__(self):
            return "Username: %s, Password: ***" % (self.username)
        
    ssh = SSHAuth()
    basic = BasicAuth()
    digest = None
    ntlm = None
    
class BaseSystem(object):
    def __init__(self, identifier):
        self.identifier = identifier
        self.authentication = Authentication()

    def inject_property(self, _property, value):
        """
        Opportunistically try to set properties.
        Assume that properties defined in configuration exist for given system.
        """
        path = _property.split('.')
        root = self.__getattribute__(path[0])
        prop = root.__getattribute__(path[1])
        prop(value)
        
    def inject_properties(self, properties):
        if properties is not None:
            for p in properties:
                key = p.keys()[0]
                self.inject_property(key, p.get(key))




# from aux.internals.pluginhook import PluginImporter
import sys
import os
import logging
import imp
import device
import service

from aux import systems_pool

log = logging.getLogger('systems')
#these imports should be kept in a system pool and only be instantiated once.

class SystemNotFoundException(Exception):pass

def scan_files(files, systemtype):
    for f in files:
        fp = open(f,'r')
        systemclass = 'class %s' % (systemtype)
        if systemclass in fp.read():
            return f
    return None

def system_filter(systems_list, blacklist):
    def filterout(item):
        for f in blacklist:
            if f in item:
                return False
        return True
    return filter(filterout, systems_list)

def find_systemtype(systemtype):
    ## NEED TO LOOKUP IN system.device and system.service and aux_device and aux_service
    ## this implementation is hack and also needs a cache file for aux devices and services so that we only need to scan plugins
    foundfile = None
    s_filter = ['ext', '.pyc']

    for module in [device, service]:
        if foundfile is None:
            modulepath = os.path.dirname(module.__file__)
            files = system_filter([os.path.join(modulepath,r) for r in os.listdir(modulepath)], s_filter)
            foundfile = scan_files(files, systemtype)

    for plugin_module in ['aux_device_', 'aux_service_']:
        if foundfile is None:
            for s in [p for p in sys.path if plugin_module in p]: 
                modulepath = imp.find_module(os.path.split(s)[1])[1]
                files = system_filter([os.path.join(modulepath,r) for r in os.listdir(modulepath)], s_filter)
                foundfile = scan_files(files, systemtype)        
        
    if foundfile is not None:
        module, fx = os.path.split(foundfile)
        mod1 = os.path.split(module)[1]
        fil1 = fx.split('.')[0]
        if 'aux_' in module:
            modlist = mod1.split('_')
            modlist.insert(1,'system')
            modlist.insert(3, 'ext')
            
        prt = imp.load_module(systemtype, open(foundfile), module, ('','',5))
        return eval("prt.%s" % systemtype)
    raise SystemNotFoundException


def get_system(systemjson):
    '''
    systemjson = {"hostname":"some.name.com.or.ip.10.0.0.1",
                  "systemtype": "MyDevice",
                  "username": "username",
                  "password": "password",
                   "properties": [
                     {"rest.authentication": [
                       {"rest.username":"rduser"},
                       {"rest.password":"yggdrasil"}
                    ]},
                     {"ssh.authentication": [
                       {"rest.username":"rduser"},
                       {"rest.password":"yggdrasil"}
                    ]}]
                 }
    '''
    
    if systemjson.get('hostname') is not None:
        #target with specific hostname        
        if systemjson.get('systemtype') is not None:
            hostname = systemjson.get('hostname')
            systemtype = systemjson.get('systemtype')
            system_instance = find_systemtype(systemtype)(hostname)
            return system_instance
        else:
            #doprobeoftype
            #TODO: this is a bit complex, the probe should be in systemdefinition
            raise NotImplementedError
    else:
        if systemjson.get('systemtype') is not None:
            for system in systems_pool:
                if systemjson.get('systemtype') == system.get('systemtype'):
                    hostname = system.get('hostname')
                    systemtype = system.get('systemtype')
                    system_instance = find_systemtype(systemtype)(hostname)
                    system_instance.inject_properties(system.get('properties'))
                    return system_instance
    return None



from setuptools import setup, find_packages
import sys, os

version = '0.0.2'

setup(name='aux',
      version=version,
      description="Automation Framework",
      long_description="""\
""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='TestCase Automation Device Network Protocol',
      author='email@protojour.com',
      author_email='',
      url='',
      license='',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'unittest2',
        'lxml',
        'nose',
        'sphinx',
        'Fabric',
        'Werkzeug',
        'colorama',
      ],
      entry_points={
        'console_scripts': [
            'aux = aux:run',
        ]},
      
      )



from unittest2 import TestCase
from aux.protocol.http.http import HTTPRequest, HTTPResponse


class HTTP_Test(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_valid_http_GET_request(self):
        request = HTTPRequest('http://localhost:7070/test', 
                              {'method':'GET',
                               'headers': {'Host': 'a.a.a',
                                           'User-Agent': 'Aux/0.1 (X11; Ubuntu; Linux x86_64; rv:24.0)',
                                           'Accept':'text/html',
                                           'Connection': 'Keep-Alive'}})
        self.assertEquals(request.method, 'GET')
        self.assertEquals(len(request.body), 0)

    def test_valid_http_POST_request(self):
        body_content = 'Hello Body'
        request = HTTPRequest('http://localhost:7070/test', 
                              {'method':'GET',
                               'headers': {'Host': 'a.a.a',
                                           'User-Agent': 'Aux/0.1 (X11; Ubuntu; Linux x86_64; rv:24.0)',
                                           'Accept':'text/html'},
                               'body': body_content})
        self.assertEquals(request.body, body_content)

    def test_default_mechanisms(self):
        request = HTTPRequest('http://localhost:7070')
        self.assertEquals(request.url.geturl(), 'http://localhost:7070/')
        self.assertEquals(request.method, 'GET')
        self.assertEquals(request.body, '')
        self.assertEquals(len(request.body), 0)

        
    def test_valid_http_200_response(self):
        response = HTTPResponse(200,
                                {'headers': {'Server': 'auxsrv/0.0.1',
                                             'Date': 'fakedate',
                                             'Connection': 'keep-alive'},
                                 'body': "200 OK"})
        self.assertEquals(response.status, 200)
        self.assertEquals(response.body, "200 OK")
                               



from aux.engine.actor.reactor import Reactor
from aux.engine.actor.proactor import Proactor
from aux.engine.actor.coactor import Coactor

__all__ = ['Reactor',
           'Proactor',
           'Coactor']

class NoActorFoundError(Exception):pass
        



import sys
import os
# import device
# import plugin
import pkg_resources

def version():
    return pkg_resources.get_distribution(aux.__package__.title()).version

def base_dir():
    return os.path.abspath(os.path.dirname(aux.__file__))

def working_dir():
    return os.getcwd()

import aux
from aux.logger import LogController
from datetime import datetime, timedelta
import json
from aux.internals import plugin_creator_routine
from aux.engine import engine_factory

logcontroller = None
configuration = None
systems_pool = []

def run():
    from aux.internals.configuration import config
    global configuration 
    global logcontroller
    global systems_pool
    configuration = config
    if config.options.plugincreator is not None:
        plugin_creator_routine(config.options.plugincreator,
                               config.args)

    ## - read config file
    try:
        config.load_default_properties()
    except Exception, e:
        print 'Falling back to default settings.'
        print e.message

    ## - initiate logger        
    logcontroller = LogController(config)

    ## - Setup
    logcontroller.summary['started'] = datetime.now()

    scripts_as_args = [script for script in config.args if '.py' in script]
    if len(scripts_as_args) != 1:
        logcontroller.runtime.error('Script argument missing')
        sys.exit(1)
    logcontroller.summary['test'] = [ sys.argv[x] for x in range(0, len(sys.argv)) if '.py' in sys.argv[x] ][0]        
    ## - initiate backend
    ## -- start engine
    engine = engine_factory('reactor', config)
    engine.start()
    ## - verify systems
    config.set_systems()
    #configuration.system
    #script
    #TODO: copy script to log_folder
    ## - run
    print(execfile(scripts_as_args[0]))
    ## - do teardown
    engine.stop()

    
# __all__ = ['device',
#            'plugin',
#            'run']

__all__ = ['run']


def exit_hook():
    if logcontroller is not None:
        logcontroller.summary['stopped'] = datetime.now()
        logcontroller.summary['runtime'] = logcontroller.summary['stopped'] - logcontroller.summary['started']
        logcontroller.pprint_summary_on_exit()
sys.exitfunc = exit_hook



from aux import working_dir
import os
import sys


def no_plugin_name_message():
    print "No plugin name"
    sys.exit(0)                

def plugin_creator_routine(plugincreator, arguments):
    # print plugincreator, arguments
    plugin_home_directory = working_dir()
    if 'service' in plugincreator:
        if len(arguments) > 0:
            packagename = "_".join(['aux', 'service', arguments[0]])
            os.system('paster create -t basic_package -o %s --no-interactive %s' % (plugin_home_directory,
                                                                                    packagename) )
        else:
            no_plugin_name_message()
    elif 'device' in plugincreator:
        if len(arguments) > 0:
            packagename = "_".join(['aux', 'device', arguments[0]])
            os.system('paster create -t basic_package -o %s --no-interactive %s' % (plugin_home_directory,
                                                                                    packagename))
        else:
            no_plugin_name_message()            
    elif 'protocol' in plugincreator:
        if len(arguments) > 0:
            packagename = "_".join(['aux', 'protocol', arguments[0]])
            os.system('paster create -t basic_package -o %s --no-interactive %s' % (plugin_home_directory,
                                                                                    packagename))
        else:
            no_plugin_name_message()            
    print 'Install plugin by running:\npip install -e %s' % packagename        
    sys.exit(0)





from unittest2 import TestCase
from aux.protocol.http import HTTP, HTTPRequest
from ..util.mockhttpserver import MockHTTPSServer

class HTTPSConnectionTest(TestCase):
    
    def setUp(self):
        self.test_server = MockHTTPSServer(port=8443)
        self.test_server.start_thread()

    def tearDown(self):
        self.test_server.stop()

    def xtest_connection_success(self):
        http = HTTP()
        http_request = HTTPRequest({'method':'GET',
                                    'headers': {'Host': 'Aux/0.1 (X11; Ubuntu; Linux x86_64; rv:24.0)',
                                                'User-Agent': 'Aux/0.1 (X11; Ubuntu; Linux x86_64; rv:24.0)',
                                                'Accept': 'text/html, application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                                                'Accept-Language': 'en-US,en;q=0.5',
                                                'Referer': 'http://abc.abc',
                                                'Cache-Control': 'max-stale=0',
                                                'Connection': 'Keep-Alive'
                                                },
                                    'data': 'fakedata'})
        response = http.send('https://127.0.0.l1:8443', http_request)
        self.assertTrue('It works!' in response)






from unittest2 import TestCase
from aux.api import http


class HTTP_API_Test(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def xtest_http_get(self):
        response = http.get(url="http://test")

    def xtest_http_post(self):
        response = http.post(url="http://test/postresource",
                             headers={})
                             
    def xtest_http_post_with_basic_authentication(self):
        response = http.post(url="http://test/postresource_with_basic_auth")

        



from aux.protocol.http.auth.basic import BasicAuthenticator


class Credentials(object):
    def __init__(self, username, password):
        self.username = username
        self.password = password


def basic(username, password):
    return BasicAuthenticator( Credentials(username, password) )








from urlparse import urlparse
from aux.api import http


class SOAPRequest(object):

    def __init__(self, url, headers, soap_content):
        self.url = url
        self.soap_ns = 'soapenv'
        self.headers = headers
        # self.soap_headers = soap_headers
        self.soap_content = soap_content
        self.body = self._request()

    def __str__(self):
        return self._request()
        
    def _request(self):
        return """<%s:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:kcen="http://kezzlerssp.com/schemas/kcengine" xmlns:com="http://kezzlerssp.com/schemas/common">
<%s:Header/>
<%s:Body>
%s
</%s:Body>
</%s:Envelope>""" % (self.soap_ns,
                     self.soap_ns,
                     self.soap_ns,
                     self.soap_content,
                     self.soap_ns,
                     self.soap_ns)

    def send(self):
        response = http.post(self.url,
                             headers=self.headers,
                             body=self.body)
        return SOAPResponse(http_response=response)

    
class SOAPResponse(object):

    def __init__(self, http_response=None):
        if http_response is not None:
            self.http_response = http_response
        self.body = self.soap2json(self.http_response.body)

    def soap2json(self, xml):
        # print xml
        
        return xml
        
    def __str__(self):
        return self.http_response.body


    

# class SOAP(object):

#     def __init__(self):
#         pass

#     def send(self, request):
#         return SOAPResponse(http.post(self.url,
#                                       header=self.headers,
#                                       body=self.body))



# request = '''
# POST https://aux.protojour.com/ws/test-ws HTTP/1.1
# Accept-Encoding: gzip,deflate
# Content-Type: text/xml;charset=UTF-8
# SOAPAction: ""
# Authorization: Basic abcdefghijklmnopqrstuvwx
# Content-Length: 352
# Host: aux.protojour.com
# Connection: Keep-Alive
# User-Agent: Apache-HttpClient/4.1.1 (java 1.5)

# <soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns="test" xmlns:com="test">
#    <soapenv:Header/>
#    <soapenv:Body>
#       <ns:ListSomething max="25">
#       </ns:ListSomething>
#    </soapenv:Body>
# </soapenv:Envelope>
# '''

# response = '''
# HTTP/1.1 200 OK
# Server: nginx/1.5.4
# Date: Wed, 12 Feb 2014 09:58:13 GMT
# Content-Type: text/xml;charset=utf-8
# Content-Length: 4734
# Connection: keep-alive
# Set-Cookie: JSESSIONID=oijoij6E0C7479C4CF531A5842241F47; Path=/; HttpOnly
# X-Request-Received: 1392199092003
# Accept: text/xml, text/html, image/gif, image/jpeg, *; q=.2, */*; q=.2
# SOAPAction: ""
# X-Src-Nginx: aux.protojour.com

# <SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/">
# <SOAP-ENV:Header/><SOAP-ENV:Body></SOAP-ENV:Body></SOAP-ENV:Envelope>
# '''
    



from socket import ( socket, AF_INET, SOCK_DGRAM,
                     IPPROTO_TCP, SOCK_STREAM,
                     SOL_SOCKET, SO_REUSEADDR )
import socket as sock
import ssl
# from ssl import wrap_socket, CERT_NONE, SSLError, PROTOCOL_SSLv23

TCP_DEFAULT_FRAME_SIZE = 1024 # power of 2 max 4096

class Transport(object):
    def __init__(self, hostname, port):
        self.addr = (hostname, port)
        self.__connection = None

    def connect(self):
        raise Exception("Not Implemented Error")
        
    def close(self):
        raise Exception("Not Implemented Error")
    
        
class UDPTransport(Transport):
    def __init__(self, hostname, port):
        super(UDPTransport, self).__init__(hostname, port)
        self.__connection = socket(AF_INET, SOCK_DGRAM)

    def connect(self):
        self.__connection.connect(self.addr)
        
    def send(self, message):
        self.__connection.sendto(message, self.addr)

    def recv(self):
        return self.__connection.recv(4096)

    def close(self):
        self.__connection.close()
    
class TCPTransport(Transport):
    def __init__(self, hostname, port, timeout=10):
        super(TCPTransport, self).__init__(hostname, port)
        self.__connection = socket(AF_INET, SOCK_STREAM)
        self.__connection.setsockopt(SOL_SOCKET,
                                     SO_REUSEADDR,
                                     1)        
    def connect(self):
        self.__connection.connect(self.addr)
        
    def send(self, message):
        self.__connection.sendto(message, self.addr)

    def recv(self, frame_size=TCP_DEFAULT_FRAME_SIZE):
        return self.__connection.recv(frame_size)
    
    def close(self):
        self.__connection.close()

    def get_hostname(self):
        return "ioctlsomething" #sock.gethostbyname(self.__connection.gethostname())
        
class CertificationHostnameMismatch(Exception):pass        

class TLS_TCPTransport(TCPTransport):
    #TODO: This should just be a wrapper 
    def __init__(self, hostname, port, timeout=0):
        super(TLS_TCPTransport, self).__init__(hostname, port)

        self.disable_ssl_certificate_validation = True
        self.timeout = timeout
        #TODO: Should do a better build up of ssl_socket
        self.__connection = ssl.wrap_socket(socket(AF_INET, SOCK_STREAM),
                                            keyfile=None,
                                            certfile=None,
                                            server_side=False,
                                            cert_reqs=ssl.CERT_NONE,
                                            ssl_version=ssl.PROTOCOL_SSLv23,
                                            ca_certs=None,
                                            do_handshake_on_connect=True,
                                            suppress_ragged_eofs=True,
                                            ciphers=None)
        self.__connection.setsockopt(SOL_SOCKET,
                                     SO_REUSEADDR,
                                     1)
        self.__connection.settimeout(self.timeout)
        
    def connect(self):
        try:
            self.__connection.connect(self.addr)
            if not self.disable_ssl_certificate_validation:
                cert = self.__connection.getpeercert()
                print cert
                hostname = "hello"
                if not self._validateCertificateHostname(cert, hostname):
                    raise CertificationHostnameMismatch(
                        'Server presented certificate that does not match ')
            
        except Exception, e:
            print e
            if self.__connection:
                self.__connection.close()
        
        
    def send(self, message):
        self.__connection.write(message)

    def recv(self, n_of_bytes=TCP_DEFAULT_FRAME_SIZE):
        try: 
            return self.__connection.read(n_of_bytes)
        except Exception, e:
            print e

    def recv_all(self):
        data = self.recv(n_of_bytes=1000)
        while data:
            print "receiving", data
            recv_buffer = data
            data = self.recv(n_of_bytes=1000)
        return recv_buffer
    
    def close(self):
        self.__connection.close()


    def _getValidHostsForCert(self, cert):
        if 'subjectAltName' in cert:
            return [x[1] for x in cert['subjectAltName']
                    if x[0].lower() == 'dns']
        else:
            return [x[0][1] for x in cert['subject']
                    if x[0][0].lower() == 'commonname']        

    def _validateCertificateHostname(self, cert, hostname):
        hosts = self._getValidHostsForCert(cert)
        for host in hosts:
            host_re = host.replace('.', '\.').replace('*', '[^.]*')
            if re.search('^%s$' % (host_re,), hostname, re.I):
                return True
        return False



from unittest2 import TestCase
from urlparse import urlparse
from aux.protocol.transport import TCPTransport, UDPTransport

from socket import (AF_INET, SOCK_DGRAM, SOL_SOCKET,
                    SO_REUSEADDR)                                         
from multiprocessing import Process
import socket
import time


class MockUDPServer(object):
    def __init__(self, port=8989):
        self.port = port
        self.host = '127.0.0.1'
        self.__socket = socket.socket(AF_INET, SOCK_DGRAM)
        self.__socket.bind((self.host, self.port))

    def start(self):
        # print "starting", self.__class__.__name__
        while True:
            d, addr = self.__socket.recvfrom(2048)
            self.__socket.sendto('response connection', addr)

    def start_thread(self):
        self.p = Process(target=self.start)
        self.p.daemon = True
        self.p.start()
        time.sleep(.01)

    def stop(self):
        self.p.terminate()
        self.__socket.close()
        
class MockTCPServer(object):
    def __init__(self, port=8989):
        self.port = port
        self.host = '127.0.0.1'#socket.gethostname()
        self.__socket = socket.socket()
        self.__socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.__socket.bind((self.host, self.port))
        
    
    def start(self):
        # print "starting", self.__class__.__name__
        self.__socket.listen(5)
        while True:
            c, addr = self.__socket.accept()
            c.send('response connection')
            c.close()

    def start_thread(self):
        self.p = Process(target=self.start)
        self.p.daemon = True
        self.p.start()
        time.sleep(.5)

    def stop(self):
        self.p.terminate()
        self.__socket.close()
        

class TCPTransportTest(TestCase):
    
    def setUp(self):
        self.test_server = MockTCPServer(port=8989)
        self.test_server.start_thread()

    def tearDown(self):
        self.test_server.stop()
        
    def test_connection(self):
        conn = TCPTransport('127.0.0.1', 8989)
        conn.connect()
        conn.send("hello world: END\n")
        data = conn.recv()
        self.assertEquals('response connection',
                          data)
        conn.close()
        

class UDPConnectionTest(TestCase):

    def setUp(self):
        self.test_server = MockUDPServer(port=8888)
        self.test_server.start_thread()

    def tearDown(self):
        self.test_server.stop()

    def test_connection(self):
        conn = UDPTransport('127.0.0.1', 8888)
        conn.connect()
        conn.send("hello world UDP: END\n")
        data = conn.recv()
        self.assertEquals('response connection',
                          data)
        conn.close()



from unittest2 import TestCase
from aux.protocol.http.http import HTTP
import struct
import os

class FakeTransport(object):

    def __init__(self, message):
        self.fake_message = message
        self.bytes_read = 0
        
    def recv(self, nofchar=1200):
        buffer = ""
        for n in xrange(0, nofchar):
            if self.bytes_read >= len(self.fake_message):
                break
            else:
                buffer += self.fake_message[self.bytes_read]
            self.bytes_read += 1
        return buffer
    
    def close(self):
        pass


class HTTPReceiveTest(TestCase):

    def test_receive_200_startline_only(self):
        message = "HTTP/1.1 200 OK\r\n"
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEquals(response.status,
                          200)

    def test_receive_200_only_headers(self):
        message = """HTTP/1.1 200 OK\r\nServer: nginx/1.5.13\r\nDate: Sat, 02 Aug 2014 19:40:38 GMT\r\nContent-Type: text/html\r\nContent-Length: 0\r\nLast-Modified: Mon, 14 Apr 2014 08:38:26 GMT\r\nConnection: keep-alive\r\nExpires: Sat, 02 Aug 2014 20:40:38 GMT\r\nCache-Control: max-age=3600\r\nAccept-Ranges: bytes\r\n\r\n"""
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEquals(len(response.body), 0)
        self.assertEquals(len(response.headers), 9)
                          
    def test_receive_200_with_json_body(self):
        message = """HTTP/1.1 200 OK\r\nContent-Type: application/json\r\nContent-Length: 15\r\n\r\n{{Hello:world}}"""
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), 15)

    def test_receive_200_with_long_body(self):
        data_length = 1664
        data = "".join(['ABCDEFGHIJKLMNOPQRSTUVWXYZ'[i%26] for i in xrange(0, data_length)])
        message = """HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %i\r\n\r\n%s""" % (data_length, data)
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), data_length)
        
    def test_receive_200_with_chunked_no_body(self):
        message = """HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nTransfer-Encoding : chunked\r\n\r\n0\r\n\r\n0"""
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), 0)

    def test_receive_200_with_chunked_no_body_one_terminating_zero(self):
        message = """HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nTransfer-Encoding : chunked\r\n\r\n0"""
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), 0)


    def test_receive_200_with_chunked_body_one_terminating_zero(self):
        content = '{"status":400,"code":"Client.UserInputException","message":"No content to map due to end-of-input\n at [Source: org.apache.catalina.connector.CoyoteInputStream@3d820d7f; line: 1, column: 1]"}'
        message = '''HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nTransfer-Encoding : chunked\r\n\r\n%s\r\n%s\r\n
0''' % (hex(len(content))[2:], content)
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len( repr(response.body)[1:-1] ), len(repr(content)[1:-1]))
        self.assertEqual(response.body,
                         content)
        
    def test_receive_200_with_chunked_body(self):
        content = """ABCDEFGHIJKLMNOPQRSTUVWXYZ"""
        message = """HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nTransfer-Encoding : chunked\r\n\r\n1a\r\n%s\r\n0\r\n\r\n0""" % content
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), 26)
        self.assertEqual(len(response.body), len(content))
        self.assertEqual(response.body, content)

    def xtest_receive_200_with_chunked_multi_body(self):
        message = """HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nTransfer-Encoding : chunked\r\n\r\n1a\r\nABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n0\r\n\r\n0"""
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), 234)

    def xtest_receive_200_with_chunked_long_body(self):
        message = """HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nTransfer-Encoding : chunked\r\n\r\n1a\r\nABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n34\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\nd0\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\nd0\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\nd0\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\nd0\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\nd0\r\nABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZABCDEFGHIJKLMNOPQRSTUVWXYZ\r\n0\r\n\r\n0"""
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), 1742)

    def xtest_receive_200_with_chunked_long_body(self):
        data_length = 4096
        data = "".join(['ABCDEFGHIJKLMNOPQRSTUVWXYZ'[i%26] for i in xrange(0, data_length)])

        message = """HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nTransfer-Encoding : chunked\r\n\r\n%s\r\n%s\r\n%s\r\n%s\r\n0\r\n\r\n0""" % (hex(data_length)[2:] ,data, hex(data_length)[2:] ,data)
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(len(response.body), data_length*2)
        
    def xtest_receive_200_with_chunked_binary_body(self):
        byte_range = 256
        data = "".join([struct.pack('B', i) for i in xrange(0,byte_range)])
        message = """HTTP/1.1 200 OK\r\nContent-Type: application/zip;charset=UTF-8\r\nTransfer-Encoding : chunked\r\nContent-Disposition : attachment; filename="test_chunkbin.zip"\r\n\r\n100\r\n%s\r\n0\r\n\r\n0""" % data
        http = HTTP()
        response = http.receive(FakeTransport(message))
        self.assertEqual(response.body, "/tmp/aux/test_chunkbin.zip")
        self.assertTrue(os.path.exists(response.body))
        self.assertEqual(os.path.getsize(response.body), byte_range)
        
        




from aux.protocol.transport import (TCPTransport, TLS_TCPTransport, TCP_DEFAULT_FRAME_SIZE)
from urlparse import urlparse, urlunparse
from aux.protocol.http.transfer import transferFactory
from aux.protocol.http.mime import mimeFactory
import auth
import logging
import aux
import re
import os
import base64

log = logging.getLogger("protocol")

uname = os.uname()
USER_AGENT = "aux/%s (%s;)" % (aux.version(),
                              " ".join([uname[0], uname[-1]]))
CRLF = "\r\n"
HTTP_DEFAULT_PORT = 80
HTTPS_DEFAULT_PORT = 443
TCP_FRAME_BUFFER_SIZE = 1500#bytes#hmmmm
#TODO: enum wrapper
HTTP_METHODS = ["OPTIONS", "GET", "HEAD",
                "POST", "PUT", "DELETE",
                "TRACE", "CONNECT"]#extension-method
# M_GET  = 'GET'
# M_POST = 'POST'
# M_PUT  = 'PUT'
# M_DELETE  = 'DELETE'
# M_HEAD = 'HEAD'

HTTP_RESPONSE_CODES = {"100": "Continue",
                       "101": "Switching Protocols",
                       "200": "OK",
                       "201": "Created",
                       "202": "Accepted",
                       "203": "Non-Authoritative Information",
                       "204": "No Content",
                       "205": "Reset Content",
                       "206": "Partial Content",
                       "300": "Multiple Choices",
                       "301": "Moved Permanently",
                       "302": "Found",
                       "303": "See Other",
                       "304": "Not Modified",
                       "305": "Use Proxy",
                       "307": "Temporary Redirect",
                       "400": "Bad Request",
                       "401": "UnauthorizedRequest",
                       "402": "Payment Required",
                       "403": "Forbidden",
                       "404": "Not Found",
                       "405": "Method Not Allowed",
                       "406": "Not Acceptable",
                       "407": "Proxy Authentication Required",
                       "408": "Request Time-out",
                       "409": "Conflict",
                       "410": "Gone",
                       "411": "Length Required",
                       "412": "Precondition Failed",
                       "413": "Request Entity Too Large",
                       "414": "Request-URI Too Large",
                       "415": "Unsupported Media Type",
                       "416": "Requested range not satisfiable",
                       "417": "Expectation Failed",
                       "500": "Internal Server Error",
                       "501": "Not Implemented",
                       "502": "Bad Gateway",
                       "503": "Service Unavailable",
                       "504": "Gateway Time-out",
                       "505": "HTTP Version not supported"}

"""
RESPONSE_HEADER = [Accept-Ranges, Age, ETag, Location, Proxy-Authenticate, Retry-After, Server, Vary, WWW-Authenticate]

"""
"""
HTTP

Authentication Type: On Request; Preemptive; SPNEGO/Kerberos; NTLM1|2

"""

class HTTPMessage(object):
    http_version = 1.1
    target = None
    headers = dict()
    body = None

    def __init__(self, headers, body):
        self.headers = headers
        self.body = body
        
    def __str__(self):
        return CRLF.join([CRLF.join([": ".join(item) for item in self.headers.items()]),
                          "",#zero-length-http-message-line
                          self.body])

class HTTPRequest(HTTPMessage):
    url = None
    def __init__(self, url, request_data={}):       
        self.method = request_data.get('method', 'GET').upper()
        self.url = urlparse(url)
        if len(self.url.path) == 0:
            l = list(self.url)
            l[2] = l[2] + "/"
            self.url = urlparse(urlunparse(l))
        headers = {'Host': self.url.hostname,
                   'User-Agent': USER_AGENT}
        headers.update(request_data.get('headers', {}))
        super(HTTPRequest, self).__init__(headers,
                                          request_data.get('body', ''))

    def __str__(self):
        path = self.url.path
        if len(self.url.query) > 0:
            path = "%s?%s" % (path, self.url.query)
        return CRLF.join(["%s %s HTTP/%0.1f" % (self.method, path, self.http_version),
                          super(HTTPRequest, self).__str__()])

    
class HTTPResponse(HTTPMessage):
    def __init__(self, status, response_data={}, request_pointer=None):
        self.status = status
        self.request_pointer = request_pointer
        super(HTTPResponse, self).__init__(response_data.get('headers', {}),
                                           response_data.get('body', ''))
        
    def __str__(self):
        return CRLF.join(["HTTP/%0.1f %s %s" % (self.http_version, self.status, HTTP_RESPONSE_CODES[str(self.status)]),
                          super(HTTPResponse, self).__str__()])
    


class HTTP(object):
    __is_persistent = False
    __has_trace = False
    
    def __init__(self):
        self.logger = logging.getLogger('protocol')
        self._transport = None
   
    def get_transport(self, url, scheme="http", persist=False, timeout=60):
        # if self._transport != None and persist:
        #     return self._transport
        #TODO: ternary default port assign is a bad idea as traceability is lost in request update url instead
        transport = None
        if "https" == scheme.lower():
            transport = TLS_TCPTransport(url.hostname,
                                         443 if url.port == None else int(url.port),
                                         timeout=timeout)
        else:
            transport = TCPTransport(url.hostname,
                                     80 if url.port == None else int(url.port),
                                     timeout=timeout)
        transport.connect()
        self.logger.debug('Connected to %s:%s' % (transport.addr[0],
                                                  transport.addr[1]))
        return transport
    
    def is_persistent(self):
        return self.__is_persistent

    def has_trace(self, should_trace=None):
        if should_trace is not None:
            self.__has_trace = should_trace
        return self.__has_trace
    
    def set_url_from_string(self, raw_url):
        url = urlparse(raw_url)
        if not url.port:
            l = list(url)
            l[1] = l[1] + ":%i" % HTTP_DEFAULT_PORT
            url = urlparse(urlunparse(l))        
        return url

    def receive(self, transport):
        #TODO: this impl needs TLC
        inbuf = transport.recv(TCP_DEFAULT_FRAME_SIZE)
        inbuf = inbuf.split("\n")
        # inbuf = transport.recv_all()
        sl = inbuf[0]
        #Validate start-line and remove it from buffer
        re_startline = re.compile(r'^HTTP\/\d\.\d\s(\d{3})\s')
        tail_msg = "\n".join(inbuf[1:])
        try:
            status = int(re_startline.match(sl).groups()[0])
        except Exception, e:
            log.error(e.message)
            raise Exception

        re_headline = re.compile(r'^([A-Za-z\-]*)\s?\:\s?(.*)')
        headers = dict()
        body = ""
        t_lines = tail_msg.split("\r\n\r\n")
        if len(t_lines) == 1: #Servers can respond with \r\n or \n
            t_lines = tail_msg.split("\n\n")
        h_lines = t_lines[0].split("\n")
        line_counter = 0
        for line in h_lines:
            line_counter += 1
            if ":" in line:
                re_group = re_headline.match(line).groups()
                headers[re_group[0]] = re_group[1]
            else:
                break
        tail_msg = tail_msg[len(t_lines[0]):]
        # if self.__has_trace:
        #     print tail_msg
        log.debug(headers)
        Transfer = transferFactory(headers)
        Mime = mimeFactory(headers)
        body = Mime(headers.get('Content-Disposition', None),
                    Transfer(headers, transport, tail_msg).read()).handle()
        response = HTTPResponse(status, {'headers': headers, 'body': body})
        log.debug("mime : %s\ntrans : %s" % (Mime, Transfer))        
        # log.debug("HTTPResponse:\n%s\n", response)
        transport.close()
        return response
    
    def send(self, request):
        request.target = request.url.hostname
        if request.method in ['POST', 'PUT', 'DELETE']:
            request.headers.update({'Content-Length': '%i' % len(request.body)})
        transport = self.get_transport(request.url, scheme=request.url.scheme)
        log.debug("HTTPRequest:\n%s\n", request)
        transport.send(str(request))
        response = self.receive(transport)
        response.request_pointer = request
        return response

class HTTPScheme(object):
    scheme = "http://"
    
    def __call__(self, scheme):
        self.scheme = scheme

    def __repr__(self):
        return self.scheme
    
class HTTPClient(object):
    auth = auth
    __has_trace = False
    scheme = HTTPScheme()
    
    def has_trace(self, should_trace=None):
        if should_trace is not None:
            self.__has_trace = should_trace
        return self.__has_trace
    
    def http_send(self, method, url, headers, body, request):
        if request == None:
            request = HTTPRequest(url,
                                  {'method': method,
                                   'headers': headers,
                                   'body': body})
        _http = HTTP()
        _http.has_trace(self.__has_trace)
        return _http.send(request)

    def head(self, url=None, headers={}, request=None):
        return self.http_send('HEAD', url, headers, "", request)
    
    def get(self, url=None, headers={}, body="", request=None):
        return self.http_send('GET', url, headers, body, request)

    def post(self, url, headers={}, body="", request=None):
        return self.http_send('POST', url, headers, body, request)

    def put(self, url, headers={}, body="", request=None):
        return self.http_send('PUT', url, headers, body, request)

    def patch(self, url, headers={}, body="", request=None):
        return self.http_send('PATCH', url, headers, body, request)
        
    def delete(self, url, headers={}, body="", request=None):
        return self.http_send('DELETE', url, headers, body, request)

    def connect(self, url, headers={}, body="", request=None):
        return self.http_send('CONNECT', url, headers, body, request)

    def trace(self, url, headers={}, body="", request=None):
        return self.http_send('TRACE', url, headers, body, request)    

    def options(self, url, headers={}, body="", request=None):
        return self.http_send('OPTIONS', url, headers, body, request)    

    def basic(self, credentials):
        return {'Authorization': 'Basic %s' % base64.b64encode(
                '%s:%s' % (credentials[0],
                           credentials[1]))}



from aux import version
from datetime import datetime
import ConfigParser, os
import logging
import shutil

class LogController(object):
    summary = dict()
    config = None
    
    def __init__(self, config):
        self.config = config
        self.loggers = dict()
        self.log_directory = os.path.join(config.options.log_directory, config.options.log_folder_name)
        self.log_console_level = config.options.log_level
        self.log_file_level = config.options.log_level
        if config.options.log_console_level is not None:
            self.log_console_level = config.options.log_console_level
        if config.options.log_file_level is not None:
            self.log_file_level = config.options.log_file_level
        self.log_verbose = config.options.verbose
        self.log_result_server = config.options.log_server
        self.log_run_dir = os.path.join(self.log_directory,
                                        datetime.strftime(datetime.now(), "%Y%m%d-%H%M%S%f"))
        if not os.path.exists(self.log_run_dir):
            os.makedirs(self.log_run_dir)
        logging.basicConfig(level=self.log_file_level,
                            format='%(asctime)s:%(name)s:%(levelname)s:%(message)s',
                            filename='%s/all.log' % self.log_run_dir,
                            filemode='w')
        for loggername in ['runtime', 'protocol', 'script', 'systems']:
            self.loggers[loggername] = self.__new_logger(loggername, self.log_run_dir)
            
        if self.log_verbose:
            self.pprint_header_on_init()
            
        self.summary['logs'] = self.log_run_dir
        self.summary['success'] = True #TODO: bad assumption
        config.options.systems = self.copy_systems_file_to_logdir(config.options.systems)
        self.summary['systems'] = config.options.systems
        
        self.runtime.debug('Config options :\n%s' % self.config.options)
        self.runtime.debug('Config arguments :\n %s' % self.config.args)

    def __getattr__(self, attr):
        if self.loggers.get(attr, None) is not None:
            return self.loggers.get(attr)
        else:
            emsg = "%s object has no attribute '%s'" % (self.__class__.__name__, attr)
            raise AttributeError(emsg)

    def __new_logger(self, loggername, logdir):
        new_logger = logging.getLogger(loggername)
        fh = logging.FileHandler(filename=os.path.join(logdir,
                                                       '%s.log' % (loggername)))
        fh.setLevel(self.log_file_level)
        ch = logging.StreamHandler()
        ch.setLevel(self.log_console_level)
        formatter = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s:%(message)s',
                                      '%H:%M:%S')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        new_logger.addHandler(fh)
        new_logger.addHandler(ch)
        new_logger.debug('Initiated')
        return new_logger
        
    def post_to_server(self):
        serverendpoint = self.config.options.log_server
        json_data = {'started' : str(self.summary.get('started')),
                     'stopped' : str(self.summary.get('stopped')),
                     'test' : self.summary.get('test'),
                     'success' : self.summary.get('success', False),
                     'testsubject' : str(self.summary.get('testsubject')),
                     'externalref': self.summary.get('externalref'),
                     'tester' : 'auxscript',
                     'logfolder' : self.summary.get('logfolder')}
        headers = {'Host': '192.168.0.135:8080', #TODO: derive from logserverpath
                   'User-Agent':'Aux/0.1 (X11;Ubuntu;Linux x86_64;rv:24.0)',
                   'Cache-Control': 'no-cache'}
        headers.update(http.basic( ('tester', 'tester'))) 
        result = http.post(serverendpoint,
                           headers=headers,
                           body=json.dumps(json_data))

    def pprint_header_on_init(self):
        if self.log_verbose:
            print "-"*70
            self.runtime.info("Options : %s" % (self.config.options))
            self.runtime.info("Args : %s" % (self.config.args))
        
    def pprint_summary_on_exit(self):
        if self.config.options.log_server is not None:
            try: 
                self.post_to_server(self.config.options.log_server)
            except:
                pass
        if self.log_verbose:
            print "-"*70
            print "- AUX %s - Summary" % version()
            print "-"*70
            for key in self.summary.keys():
                print "- %s: %s" % (key, self.summary[key])
            print "-"*70
                
    def copy_systems_file_to_logdir(self, systems):
        newurl = None
        if systems is not None:
            if '.json' in systems:
                url = os.path.abspath(systems)
                newurl = os.path.join(self.log_run_dir, systems)
                shutil.copyfile(url, newurl)
            else:
                newurl = os.path.join(self.log_run_dir, "systems.json")
                fp = open(newurl,"w")
                fp.write(systems)
                fp.close()
            return newurl
        return "No systems set."

        

        
